/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "planner.h"

#include <memory>

#include "execution/executor_delete.h"
#include "execution/executor_index_scan.h"
#include "execution/executor_insert.h"
#include "execution/executor_nestedloop_join.h"
#include "execution/executor_projection.h"
#include "execution/executor_seq_scan.h"
#include "execution/executor_update.h"
#include "index/ix.h"
#include "record_printer.h"

// 目前的索引匹配规则为：完全匹配索引字段，且全部为单点查询，不会自动调整where条件的顺序
bool Planner::get_index_cols(std::string tab_name, std::vector<Condition> &curr_conds, std::vector<std::string>& index_col_names) {
    index_col_names.clear();
    // // 这里筛选出condition条件，仅筛选出右侧为值且表正确的condition
    // for(auto& cond: curr_conds) {
    //     // if(cond.is_rhs_val && cond.op == OP_EQ && cond.lhs_col.tab_name.compare(tab_name) == 0)
    //     //     index_col_names.push_back(cond.lhs_col.col_name);
    //     if(cond.is_rhs_val && cond.lhs_col.tab_name.compare(tab_name) == 0)
    //         index_col_names.push_back(cond.lhs_col.col_name);
    // }
    // 索引的匹配规则在这里重写
    TabMeta& tab = sm_manager_->db_.get_table(tab_name);
    std::vector<int> match_res(tab.indexes.size(), 0);
    for(size_t i=0; i<tab.indexes.size(); i++){
        auto &index = tab.indexes[i];
        bool end = false;
        for(int j=0; j<index.col_num; j++){
            bool found = false;
            for(auto &con: curr_conds){
                if(index.cols[j].name == con.lhs_col.col_name && con.is_rhs_val && con.op == OP_EQ && con.lhs_col.tab_name.compare(tab_name) == 0){
                    match_res[i]++;
                    found = true;
                    break;
                }
                else if(index.cols[j].name == con.lhs_col.col_name && con.is_rhs_val && con.op != OP_NE && con.lhs_col.tab_name.compare(tab_name) == 0){
                    // 范围查找，则匹配到此结束
                    match_res[i]++;
                    found = true;
                    end = true;
                    break;
                }
            }
            if(!found || end) break;
        }
    }
    auto max_match = std::max_element(match_res.begin(), match_res.end());
    if(tab.indexes.size()==0 || max_match == match_res.end() || *max_match == 0)
        return false;
    // 找到最大匹配索引，重新调整condition顺序
    auto &index = tab.indexes[std::distance(match_res.begin(), max_match)];
    for(auto col: index.cols){
        index_col_names.push_back(col.name);
    }
    int off = 0;
    for(int i=0; i<*max_match; i++){
        for(size_t j = off; j<curr_conds.size(); j++){
            if(curr_conds[j].lhs_col.col_name == index.cols[i].name){
                std::swap(curr_conds[off], curr_conds[j]);
                off++;
            }
        }
    }
    // if(tab.is_index(index_col_names)) return true;
    return true;
}

/**
 * @brief 表算子条件谓词生成
 *
 * @param conds 条件
 * @param tab_names 表名
 * @return std::vector<Condition>
 */
std::vector<Condition> pop_conds(std::vector<Condition> &conds, std::string tab_names) {
    // auto has_tab = [&](const std::string &tab_name) {
    //     return std::find(tab_names.begin(), tab_names.end(), tab_name) != tab_names.end();
    // };
    std::vector<Condition> solved_conds;
    auto it = conds.begin();
    while (it != conds.end()) {
        if ((tab_names.compare(it->lhs_col.tab_name) == 0 && it->is_rhs_val) || (it->lhs_col.tab_name.compare(it->rhs_col.tab_name) == 0)) {
            solved_conds.emplace_back(std::move(*it));
            it = conds.erase(it);
        } else {
            it++;
        }
    }
    return solved_conds;
}

int push_conds(Condition *cond, std::shared_ptr<Plan> plan)
{
    if(auto x = std::dynamic_pointer_cast<ScanPlan>(plan))
    {
        if(x->tab_name_.compare(cond->lhs_col.tab_name) == 0) {
            return 1;
        } else if(x->tab_name_.compare(cond->rhs_col.tab_name) == 0){
            return 2;
        } else {
            return 0;
        }
    }
    else if(auto x = std::dynamic_pointer_cast<JoinPlan>(plan))
    {
        int left_res = push_conds(cond, x->left_);
        // 条件已经下推到左子节点
        if(left_res == 3){
            return 3;
        }
        int right_res = push_conds(cond, x->right_);
        // 条件已经下推到右子节点
        if(right_res == 3){
            return 3;
        }
        // 左子节点或右子节点有一个没有匹配到条件的列
        if(left_res == 0 || right_res == 0) {
            return left_res + right_res;
        }
        // 左子节点匹配到条件的右边
        if(left_res == 2) {
            // 需要将左右两边的条件变换位置
            std::map<CompOp, CompOp> swap_op = {
                {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
            };
            std::swap(cond->lhs_col, cond->rhs_col);
            cond->op = swap_op.at(cond->op);
        }
        x->conds_.emplace_back(std::move(*cond));
        return 3;
    }
    return false;
}

std::shared_ptr<Plan> pop_scan(int *scantbl, std::string table, std::vector<std::string> &joined_tables, 
                std::vector<std::shared_ptr<Plan>> plans)
{   // 用于产生连接的plan, 其作用是: 在scan plans中找到需要join的table
    // 将table加入到joined_tables中, 并将scantbl标记为1, 最后返回scan算子
    for (size_t i = 0; i < plans.size(); i++) {
        auto x = std::dynamic_pointer_cast<ScanPlan>(plans[i]);
        if(x->tab_name_.compare(table) == 0)
        {
            scantbl[i] = 1;
            joined_tables.emplace_back(x->tab_name_);
            return plans[i];
        }
    }
    return nullptr;
}


std::shared_ptr<Query> Planner::logical_optimization(std::shared_ptr<Query> query, Context *context)
{
    
    //TODO 实现逻辑优化规则

    return query;
}

std::shared_ptr<Plan> Planner::physical_optimization(std::shared_ptr<Query> query, Context *context)
{
    std::shared_ptr<Plan> plan = make_one_rel(query);
    
    // 其他物理优化

    // 处理order by
    plan = generate_sort_plan(query, std::move(plan)); 

    return plan;
}


// 根据query, 如果只涉及一个表, 则返回scanExecutor
// 如果涉及多个表, 则返回join_executor
std::shared_ptr<Plan> Planner::make_one_rel(std::shared_ptr<Query> query)
{
    auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse);
    std::vector<std::string> tables = query->tables;
    // Scan table , 生成表算子列表tab_nodes
    std::vector<std::shared_ptr<Plan>> table_scan_executors(tables.size());
    for (size_t i = 0; i < tables.size(); i++) {
        auto curr_conds = pop_conds(query->conds, tables[i]);
        // int index_no = get_indexNo(tables[i], curr_conds);
        std::vector<std::string> index_col_names;
        bool index_exist = get_index_cols(tables[i], curr_conds, index_col_names);
        if (index_exist == false) {  // 该表没有索引
            index_col_names.clear();
            table_scan_executors[i] = std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, tables[i], curr_conds, index_col_names);
        } else {  // 存在索引
            table_scan_executors[i] = std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, tables[i], curr_conds, index_col_names);
        }
    }
    // 只有一个表，不需要join。
    if(tables.size() == 1)
    {
        return table_scan_executors[0];
    }
    // 获取where条件
    auto conds = std::move(query->conds);
    std::shared_ptr<Plan> table_join_executors;
    
    int scantbl[tables.size()];
    for(size_t i = 0; i < tables.size(); i++)
        scantbl[i] = -1;
    
    // 假设在ast中已经添加了jointree，这里需要修改的逻辑是，先处理jointree，然后再考虑剩下的部分
    if(conds.size() >= 1) // 有连接条件
    {
        // 根据连接条件，生成第一层join
        std::vector<std::string> joined_tables(tables.size());
        auto it = conds.begin();
        if(it != conds.end())
        {
            std::shared_ptr<Plan> left , right;
            left = pop_scan(scantbl, it->lhs_col.tab_name, joined_tables, table_scan_executors);
            right = pop_scan(scantbl, it->rhs_col.tab_name, joined_tables, table_scan_executors);
            std::vector<Condition> join_conds{*it};
            std::string left_tab_name(it->lhs_col.tab_name), right_table_name(it->rhs_col.tab_name);
            it = conds.erase(it);
            bool can_hash_join = false;
            for(size_t i=0;i<conds.size();i++)
            {
                auto iter_cond = conds[i];
                if(iter_cond.lhs_col.tab_name == left_tab_name && 
                   iter_cond.rhs_col.tab_name == right_table_name)
                {
                    if(it->op == OP_EQ)  // 如果存在等值条件, 则可以使用hash join
                        can_hash_join = true;
                    join_conds.push_back(iter_cond);
                    it = conds.erase(conds.begin()+i);  // 这个会不会导致错位??
                }
            }
            if(can_hash_join)
                table_join_executors = std::make_shared<JoinPlan>(T_Hash, std::move(left), std::move(right), join_conds);
            else
                table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(left), std::move(right), join_conds);
        }
        // while (it != conds.end()) {
        //     std::shared_ptr<Plan> left , right;
        //     left = pop_scan(scantbl, it->lhs_col.tab_name, joined_tables, table_scan_executors);
        //     right = pop_scan(scantbl, it->rhs_col.tab_name, joined_tables, table_scan_executors);
        //     std::vector<Condition> join_conds{*it};
        //     if(it->lhs_col.tab_name == it->lhs_col.tab_name)
        //     //建立join
        //     table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(left), std::move(right), join_conds);
        //     it = conds.erase(it);
        //     break;
        // }
        // 根据连接条件，生成第2-n层join
        it = conds.begin();
        while (it != conds.end()) {
            std::shared_ptr<Plan> left_need_to_join_executors = nullptr;
            std::shared_ptr<Plan> right_need_to_join_executors = nullptr;
            bool isneedreverse = false;
            // 如果当前条件使用的condition左侧table不在joined_tables中, 说明这个table还没有被连接, 需要继续生成join算子
            if (std::find(joined_tables.begin(), joined_tables.end(), it->lhs_col.tab_name) == joined_tables.end()) {
                left_need_to_join_executors = pop_scan(scantbl, it->lhs_col.tab_name, joined_tables, table_scan_executors);
            }

            // 如果当前条件使用的condition右侧table不在joined_tables中, 说明这个table还没有被连接, 需要继续生成join算子
            if (std::find(joined_tables.begin(), joined_tables.end(), it->rhs_col.tab_name) == joined_tables.end()) {
                right_need_to_join_executors = pop_scan(scantbl, it->rhs_col.tab_name, joined_tables, table_scan_executors);
                isneedreverse = true;
            } 

            // 如果两个都不为空, 说明当前条件的左右两个table需要先连接一下, 连接后, 再与其他的连接结果再进行连接
            if(left_need_to_join_executors != nullptr && right_need_to_join_executors != nullptr) {
                std::vector<Condition> join_conds{*it}; // 这个目前好像只支持单个条件

                std::shared_ptr<Plan> temp_join_executors = nullptr;
                if(it->op == OP_EQ)
                    temp_join_executors = std::make_shared<JoinPlan>(T_Hash, 
                                                                        std::move(left_need_to_join_executors), 
                                                                        std::move(right_need_to_join_executors), 
                                                                        join_conds);
                else                                                
                    temp_join_executors = std::make_shared<JoinPlan>(T_NestLoop, 
                                                                        std::move(left_need_to_join_executors), 
                                                                        std::move(right_need_to_join_executors), 
                                                                        join_conds);

                table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(temp_join_executors), 
                                                                    std::move(table_join_executors), 
                                                                    std::vector<Condition>());

            // 如果只有一个为空, 说明其中一个table需要连接
            // 如果是右侧的table需要连接, 那么condition需要先反转reverse一下, 再添加到plan里面
            // 如果是左侧, 那么直接添加就行
            } else if(left_need_to_join_executors != nullptr || right_need_to_join_executors != nullptr) {
                if(isneedreverse) {
                    std::map<CompOp, CompOp> swap_op = {
                        {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
                    };
                    std::swap(it->lhs_col, it->rhs_col);
                    it->op = swap_op.at(it->op);
                    left_need_to_join_executors = std::move(right_need_to_join_executors);
                }
                std::vector<Condition> join_conds{*it};
                if(it->op == OP_EQ)               
                    table_join_executors = std::make_shared<JoinPlan>(T_Hash, std::move(left_need_to_join_executors), 
                                                                        std::move(table_join_executors), join_conds);
                else
                    table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(left_need_to_join_executors), 
                                                                        std::move(table_join_executors), join_conds);
            } else { // 如果都为空, 爱咋咋地吧
                push_conds(std::move(&(*it)), table_join_executors);
            }
            it = conds.erase(it);
        }
    } else {  // 妹有连接条件
        table_join_executors = table_scan_executors[0];
        scantbl[0] = 1;
    }

    //连接剩余表
    for (size_t i = 0; i < tables.size(); i++) {
        if(scantbl[i] == -1) {
            table_join_executors = std::make_shared<JoinPlan>(T_NestLoop, std::move(table_scan_executors[i]), std::move(table_join_executors), std::vector<Condition>());
        }
    }

    return table_join_executors;
}


std::shared_ptr<Plan> Planner::generate_sort_plan(std::shared_ptr<Query> query, std::shared_ptr<Plan> plan)
{
    auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse);
    if(!x->has_sort) {
        return plan;
    }
    std::vector<std::string> tables = query->tables;
    std::vector<ColMeta> all_cols;
    for (auto &sel_tab_name : tables) {
        // 这里db_不能写成get_db(), 注意要传指针
        const auto &sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
    }
    std::vector<ColMeta> sort_cols;  // 这个sort_cols是排序的column
    for(auto& x_col:x->order->cols) {
        for (auto &col : all_cols)
        {
            // 如果tabname 不为空, 那么tabname和colname必须都匹配
            // 如果tabname 为空, 那么只用colname匹配, tabname为当前colname
            // 该逻辑正确的前提是: 用来排序的属性列名字是唯一的, 不存在两个表有同名属性列
            if(x_col->col_name == col.name)
            {
                if(x_col->tab_name.empty())
                    sort_cols.push_back(std::move(col));
                else
                {
                    if(x_col->tab_name == col.tab_name)
                        sort_cols.push_back(std::move(col));
                }
            }
        }
    }
    // std::reverse(sort_cols.begin(),sort_cols.end()); // 之前是逆序压栈的, 现在翻转回来
    std::vector<bool> is_desc_arr;
    for(auto is_desc:x->order->orderby_dir)
    {   // 只要不是降序, 默认和升序都是一样的
        is_desc_arr.push_back(is_desc == ast::OrderBy_DESC);
    }
    return std::make_shared<SortPlan>(T_Sort, std::move(plan), sort_cols, is_desc_arr);
}


/**
 * @brief select plan 生成
 *
 * @param sel_cols select plan 选取的列
 * @param tab_names select plan 目标的表
 * @param conds select plan 选取条件
 */
std::shared_ptr<Plan> Planner::generate_select_plan(std::shared_ptr<Query> query, Context *context) {
    //逻辑优化
    query = logical_optimization(std::move(query), context);

    //物理优化
    auto sel_cols = query->cols;
    std::shared_ptr<Plan> plannerRoot = physical_optimization(query, context);
    plannerRoot = std::make_shared<ProjectionPlan>(T_Projection, std::move(plannerRoot), std::move(sel_cols));

    return plannerRoot;
}

// 生成DDL语句和DML语句的查询执行计划
std::shared_ptr<Plan> Planner::do_planner(std::shared_ptr<Query> query, Context *context)
{
    std::shared_ptr<Plan> plannerRoot;
    if (auto x = std::dynamic_pointer_cast<ast::CreateTable>(query->parse)) {
        // create table;
        std::vector<ColDef> col_defs;
        for (auto &field : x->fields) {
            if (auto sv_col_def = std::dynamic_pointer_cast<ast::ColDef>(field)) {
                ColDef col_def = {.name = sv_col_def->col_name,
                                  .type = interp_sv_type(sv_col_def->type_len->type),
                                  .len = sv_col_def->type_len->len};
                col_defs.push_back(col_def);
            } else {
                throw InternalError("Unexpected field type");
            }
        }
        plannerRoot = std::make_shared<DDLPlan>(T_CreateTable, x->tab_name, std::vector<std::string>(), col_defs);
    } else if (auto x = std::dynamic_pointer_cast<ast::DropTable>(query->parse)) {
        // drop table;
        plannerRoot = std::make_shared<DDLPlan>(T_DropTable, x->tab_name, std::vector<std::string>(), std::vector<ColDef>());
    } else if (auto x = std::dynamic_pointer_cast<ast::CreateIndex>(query->parse)) {
        // create index;
        plannerRoot = std::make_shared<DDLPlan>(T_CreateIndex, x->tab_name, x->col_names, std::vector<ColDef>());
    } else if (auto x = std::dynamic_pointer_cast<ast::DropIndex>(query->parse)) {
        // drop index
        plannerRoot = std::make_shared<DDLPlan>(T_DropIndex, x->tab_name, x->col_names, std::vector<ColDef>());
    } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(query->parse)) {
        // insert;
        plannerRoot = std::make_shared<DMLPlan>(T_Insert, std::shared_ptr<Plan>(), x->tab_name, query->values, std::vector<Condition>(), std::vector<SetClause>());
    } else if (auto x = std::dynamic_pointer_cast<ast::LoadStmt>(query->parse)) {
        // Load;
        plannerRoot = std::make_shared<DMLPlan>(T_Insert, std::shared_ptr<Plan>(), x->tab_name, std::vector<Value>(), std::vector<Condition>(), std::vector<SetClause>(), x->file_name);
    } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(query->parse)) {
        // delete;
        // 生成表扫描方式
        std::shared_ptr<Plan> table_scan_executors;
        // 只有一张表，不需要进行物理优化了
        // int index_no = get_indexNo(x->tab_name, query->conds);
        std::vector<std::string> index_col_names;
        bool index_exist = get_index_cols(x->tab_name, query->conds, index_col_names);
        
        if (index_exist == false) {  // 该表没有索引
            index_col_names.clear();
            table_scan_executors = 
                std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name, query->conds, index_col_names);
        } else {  // 存在索引
            table_scan_executors =
                std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name, query->conds, index_col_names);
        }

        plannerRoot = std::make_shared<DMLPlan>(T_Delete, table_scan_executors, x->tab_name,  
                                                std::vector<Value>(), query->conds, std::vector<SetClause>());
    } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(query->parse)) {
        // update;
        // 生成表扫描方式
        std::shared_ptr<Plan> table_scan_executors;
        // 只有一张表，不需要进行物理优化了
        // int index_no = get_indexNo(x->tab_name, query->conds);
        std::vector<std::string> index_col_names;
        bool index_exist = get_index_cols(x->tab_name, query->conds, index_col_names);

        if (index_exist == false) {  // 该表没有索引
            index_col_names.clear();
            table_scan_executors = 
                std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name, query->conds, index_col_names);
        } else {  // 存在索引
            table_scan_executors =
                std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name, query->conds, index_col_names);
        }
        plannerRoot = std::make_shared<DMLPlan>(T_Update, table_scan_executors, x->tab_name,
                                                     std::vector<Value>(), query->conds, 
                                                     query->set_clauses);
    } else if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse)) {

        std::shared_ptr<plannerInfo> root = std::make_shared<plannerInfo>(x);
        // 生成select语句的查询执行计划
        std::shared_ptr<Plan> projection = generate_select_plan(std::move(query), context);
        plannerRoot = std::make_shared<DMLPlan>(T_select, projection, std::string(), std::vector<Value>(), std::vector<Condition>(), std::vector<SetClause>());
    } else {
        throw InternalError("Unexpected AST root");
    }
    return plannerRoot;
}