/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "lru_k_replacer.h"

LRUKReplacer::LRUKReplacer(size_t num_pages, int k) : max_size_(num_pages), k_(k) {}

LRUKReplacer::~LRUKReplacer() {}

bool LRUKReplacer::victim(frame_id_t *frame_id) 
{
    std::lock_guard<std::mutex> lock(latch_);
    if (!history_list_.empty()) {
        *frame_id = history_list_.back();
        history_list_.pop_back();
        history_hash_.erase(*frame_id);
        use_count_map_.erase(*frame_id);
        return true;
    } else if (!LRUKlist_.empty()) {
        *frame_id = LRUKlist_.back();
        LRUKlist_.pop_back();
        LRUKhash_.erase(*frame_id);
        use_count_map_.erase(*frame_id);
        return true;
    } else {
        frame_id = nullptr;
        return false;
    }
}

void LRUKReplacer::pin(frame_id_t frame_id) 
{
    std::lock_guard<std::mutex> lock(latch_);

    auto it = LRUKhash_.find(frame_id);
    if (it != LRUKhash_.end()) {
        LRUKlist_.erase(it->second);
        LRUKhash_.erase(it);
        use_count_map_.erase(frame_id);
        return;
    }

    it = history_hash_.find(frame_id);
    if (it != history_hash_.end()) {
        history_list_.erase(it->second);
        history_hash_.erase(it);
        use_count_map_.erase(frame_id);
        return;
    }
    
    return;
}

void LRUKReplacer::unpin(frame_id_t frame_id) 
{
    std::lock_guard<std::mutex> lock(latch_);  // 加锁

    auto it = LRUKhash_.find(frame_id);  // 先在lru-k缓存队列中找
    if (it != LRUKhash_.end()) {  // 如果存在，就重新移到队首
        LRUKlist_.erase(it->second);
        LRUKlist_.push_front(frame_id);
        LRUKhash_[frame_id] = LRUKlist_.begin();
        use_count_map_[frame_id] = k_;
    } else {  // 如果不在lru-k缓存队列
        it = history_hash_.find(frame_id);  // 就去history队列中找
        if (it != history_hash_.end()) {  // 如果存在，就增加计数
            use_count_map_[frame_id]++;
            if(use_count_map_[frame_id] == k_) {  // 如果计数达到k，就从history队列移动到lru-k队列。
                history_list_.erase(it->second);
                history_hash_.erase(it);
                LRUKlist_.push_front(frame_id);
                LRUKhash_[frame_id] = LRUKlist_.begin();
            }
        } else {  // 如果不在history队列，就加到history队列
            history_list_.push_front(frame_id);
            history_hash_[frame_id] = history_list_.begin();
            use_count_map_[frame_id] = 1;
        }
    }
    return;
}

size_t LRUKReplacer::Size() {
    std::lock_guard<std::mutex> lock(latch_);
    return history_list_.size() + LRUKlist_.size();
}





